from .base import Atomic, AtomicView, AtomicViewContext
from .core import AtomicCore

from .mixins.types import BYTES


class AtomicBytes(Atomic, BYTES):

    def __init__(self, *, width: int):
        super().__init__(width=width, is_integral=False, is_signed=False)


class AtomicBytesView(AtomicView, BYTES):

    def __init__(self, core: AtomicCore):
        super().__init__(core)


class AtomicBytesViewContext(AtomicViewContext):

    def __init__(self, *, buffer):
        super().__init__(buffer=buffer, is_integral=False, is_signed=False)

    def __enter__(self) -> AtomicBytesView:
        self._assert_enter_preconditions()
        self._entered = True
        return AtomicBytesView(self._core)



from .exceptions import UnsupportedWidthException
from .patomic import Patomic
from .pybuffer import PyBuffer


class Alignment:

    def __init__(self, width: int):
        # check that the width is supported
        p = Patomic()
        ops = p.ops(width)
        if p.count_nonnull_ops(ops, readonly=False) == 0:
            raise UnsupportedWidthException(width, readonly=False)
        # get alignment info
        align = p.alignment(width)
        self.width: int = width
        self.recommended: int = align.recommended
        self.minimum: int = align.minimum
        self.size_within: int = align.size_within

    def __str__(self):
        msg = f"{self.__class__.__name__}(width={self.width}, " \
              f"recommended={self.recommended}, minimum=" \
              f"{self.minimum}, size_within={self.size_within})"
        return msg

    def _checked_buffer(self, buffer) -> PyBuffer:
        # check support for buffer protocol
        try:
            pybuf = PyBuffer(buffer, writeable=False)
            # check width
            if pybuf.width != self.width:
                error_msg = "Positional argument 'buffer' does not have" \
                            f" matching width of {self.width}."
                raise ValueError(error_msg)
            # return for use in validation functions
            return pybuf
        except TypeError:
            pass
        # caught TypeError (raise here instead of within except context)
        error_msg = "Positional argument 'buffer' must support the " \
                    "buffer protocol."
        raise TypeError(error_msg)

    def is_valid(self, buffer, *, using_recommended: bool = True) -> bool:
        with self._checked_buffer(buffer) as pybuf:
            # recommended
            if using_recommended:
                return (pybuf.address % self.recommended) == 0
            # minimum
            elif (pybuf.address % self.minimum) == 0:
                if self.size_within == 0:
                    return True
                else:
                    address = pybuf.address % self.size_within
                    return (address + self.width) <= self.size_within
            # no support
            else:
                return False

    def is_valid_recommended(self, buffer) -> bool:
        return self.is_valid(buffer, using_recommended=True)

    def is_valid_minimum(self, buffer) -> bool:
        return self.is_valid(buffer, using_recommended=False)



from ...enums import MemoryOrder, OpType

from ..core import AtomicCore

from .byteops import ByteOperationsMixin
from .cmpxchg import CmpxchgResult

import sys
from typing import Optional, Tuple


class _ImplIntegralOperationsMixin(ByteOperationsMixin):

    _core: AtomicCore

    def _impl_cmpxchg(self, optype: OpType, expected: int, desired: int,
                      succ: MemoryOrder, fail: MemoryOrder) -> CmpxchgResult[int]:
        expected = expected.to_bytes(self._core.width, sys.byteorder, signed=self._core.signed)
        desired = desired.to_bytes(self._core.width, sys.byteorder, signed=self._core.signed)
        ok, exp = super()._impl_cmpxchg(optype, expected, desired, succ, fail)
        exp = int.from_bytes(exp, sys.byteorder, signed=self._core.signed)
        return CmpxchgResult(ok, exp)

    def _impl_bin_ari(self, optype: OpType, value: Optional[int],
                      order: MemoryOrder) -> Optional[int]:
        if value is not None:
            value = value.to_bytes(self._core.width, sys.byteorder, signed=self._core.signed)
        res = super()._impl_bin_ari(optype, value, order)
        if res is not None:
            res = int.from_bytes(res, sys.byteorder, signed=self._core.signed)
        return res


class IntegralOperationsMixin(_ImplIntegralOperationsMixin):

    _core: AtomicCore

    def store(self, desired: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        desired = desired.to_bytes(self._core.width, sys.byteorder, signed=self._core.signed)
        super().store(desired, order)

    def load(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        value = super().load(order)
        return int.from_bytes(value, sys.byteorder, signed=self._core.signed)

    def exchange(self, desired: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        desired = desired.to_bytes(self._core.width, sys.byteorder, signed=self._core.signed)
        value = super().exchange(desired, order)
        return int.from_bytes(value, sys.byteorder, signed=self._core.signed)

    def cmpxchg_weak(self, expected: int, desired: int,
                     success: MemoryOrder = MemoryOrder.SEQ_CST,
                     failure: MemoryOrder = MemoryOrder.SEQ_CST) -> CmpxchgResult[int]:
        return self._impl_cmpxchg(OpType.CMPXCHG_WEAK, expected, desired, success, failure)

    def cmpxchg_strong(self, expected: int, desired: int,
                       success: MemoryOrder = MemoryOrder.SEQ_CST,
                       failure: MemoryOrder = MemoryOrder.SEQ_CST) -> CmpxchgResult[int]:
        return self._impl_cmpxchg(OpType.CMPXCHG_STRONG, expected, desired, success, failure)

    def bin_or(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.OR, value, order)

    def bin_xor(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.XOR, value, order)

    def bin_and(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.AND, value, order)

    def bin_not(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.NOT, None, order)

    def bin_fetch_or(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_OR, value, order)

    def bin_fetch_xor(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_XOR, value, order)

    def bin_fetch_and(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_AND, value, order)

    def bin_fetch_not(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_NOT, None, order)

    def add(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.ADD, value, order)

    def sub(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.SUB, value, order)

    def inc(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.INC, None, order)

    def dec(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.DEC, None, order)

    def neg(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.NEG, None, order)

    def fetch_add(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_ADD, value, order)

    def fetch_sub(self, value: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_SUB, value, order)

    def fetch_inc(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_INC, None, order)

    def fetch_dec(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_DEC, None, order)

    def fetch_neg(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> int:
        return self._impl_bin_ari(OpType.FETCH_NEG, None, order)



from .base import Atomic, AtomicView, AtomicViewContext
from .core import AtomicCore

from .mixins.types import INTEGRAL


class AtomicIntegral(Atomic, INTEGRAL):

    def __init__(self, *, width: int, is_signed: bool):
        super().__init__(width=width, is_integral=True, is_signed=is_signed)


class AtomicIntegralView(AtomicView, INTEGRAL):

    def __init__(self, core: AtomicCore):
        super().__init__(core)


class AtomicIntegralViewContext(AtomicViewContext):

    def __init__(self, *, buffer, is_signed: bool):
        super().__init__(buffer=buffer, is_integral=True, is_signed=is_signed)

    def __enter__(self)-> AtomicIntegralView:
        self._assert_enter_preconditions()
        self._entered = True
        return AtomicIntegralView(self._core)



import cffi


ffi = cffi.FFI()


class PyBuffer:

    def __init__(self, exporter, *, writeable: bool, force: bool = False):
        # check if __init__ has been called
        if hasattr(self, "_buf"):
            raise ValueError("PyBuffer object cannot be re-initialised.")
        # check if writeable can be satisfied
        with memoryview(exporter) as view:
            if writeable and view.readonly and not force:
                raise RuntimeError("Cannot create writeable PyBuffer from readonly exporter.")
            # get and save buffer
            self._buf = ffi.from_buffer("char[]", exporter, not view.readonly)
            self._obj = exporter
            self._len = view.nbytes
            self._readonly = not writeable

    def __enter__(self):
        self._assert_not_released()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __del__(self):
        self.release()

    def __bool__(self):
        return not self._released

    def __str__(self):
        if self:
            return f"{self.__class__.__name__}(address={self.address}, width={self.width}" \
                   f", readonly={self.readonly}, obj={self.obj})"
        else:
            return f"{self.__class__.__name__}(released)"

    def release(self) -> None:
        # this may be called in __del__ if exception is raised in __init__
        # we cannot rely on any attributes existing
        if hasattr(self, "_buf") and self._buf is not None:
            # if _buf exists, we can be confident the rest of the attributes do too
            ffi.release(self._buf)
            self._buf = None
            self._obj = None
            self._len = None
            self._readonly = None

    @property
    def _released(self) -> bool:
        return self._buf is None

    def _assert_not_released(self) -> None:
        if self._released:
            msg = f"Operation forbidden on released {self.__class__.__name__} object."
            raise ValueError(msg)

    @property
    def address(self) -> int:
        self._assert_not_released()
        return int(ffi.cast("uintptr_t", self._buf))

    @property
    def width(self) -> int:
        self._assert_not_released()
        return self._len

    @property
    def readonly(self) -> bool:
        self._assert_not_released()
        return self._readonly

    @property
    def obj(self):
        self._assert_not_released()
        return self._obj



from ...enums import OpType
from ...exceptions import UnsupportedOperationException

from ..core import AtomicCore

from typing import Callable


class BasePropertiesMixin:

    _core: AtomicCore

    @property
    def _address(self) -> int:
        return self._core.address

    @property
    def width(self) -> int:
        return self._core.width

    @property
    def readonly(self) -> bool:
        return self._core.readonly

    @property
    def ops_supported(self) -> [OpType]:
        return self._core.ops_supported

    def __str__(self):
        msg = f"{self.__class__.__name__}(width={self.width}, " \
              f"readonly={self.readonly})"
        return msg


class BytePropertiesMixin(BasePropertiesMixin):

    load: Callable[[], bytes]

    def __str__(self):
        try:
            value = self.load()
        except UnsupportedOperationException:
            value = "LOAD_NOT_SUPPORTED"
        msg = f"{self.__class__.__name__}(value={value}, " \
              f"width={self.width}, readonly={self.readonly})"
        return msg

    def __bytes__(self):
        return self.load()


class IntegralPropertiesMixin(BasePropertiesMixin):

    load: Callable[[], int]

    def __str__(self):
        try:
            value = self.load()
        except UnsupportedOperationException:
            value = "LOAD_NOT_SUPPORTED"
        msg = f"{self.__class__.__name__}(value={value}, " \
              f"width={self.width}, readonly={self.readonly}, " \
              f"signed={self.signed})"
        return msg

    def __int__(self):
        return self.load()

    @property
    def signed(self) -> bool:
        return self._core.signed



from .baseint import AtomicIntegral, AtomicIntegralView, AtomicIntegralViewContext
from .core import AtomicCore

from .mixins.types import INT, UINT


class AtomicInt(AtomicIntegral, INT):

    def __init__(self, *, width: int):
        super().__init__(width=width, is_signed=True)


class AtomicIntView(AtomicIntegralView, INT):

    def __init__(self, core: AtomicCore):
        super().__init__(core)


class AtomicIntViewContext(AtomicIntegralViewContext):

    def __init__(self, *, buffer):
        super().__init__(buffer=buffer, is_signed=True)

    def __enter__(self) -> AtomicIntView:
        self._assert_enter_preconditions()
        self._entered = True
        return AtomicIntView(self._core)


class AtomicUint(AtomicIntegral, UINT):

    def __init__(self, *, width: int):
        super().__init__(width=width, is_signed=False)


class AtomicUintView(AtomicIntegralView, UINT):

    def __init__(self, core: AtomicCore):
        super().__init__(core)


class AtomicUintViewContext(AtomicIntegralViewContext):

    def __init__(self, *, buffer):
        super().__init__(buffer=buffer, is_signed=False)

    def __enter__(self) -> AtomicUintView:
        self._assert_enter_preconditions()
        self._entered = True
        return AtomicUintView(self._core)



import pathlib

from ctypes import *


_opsig_explicit_store_t = CFUNCTYPE(None, c_void_p, c_void_p, c_int)
_opsig_explicit_load_t = CFUNCTYPE(None, c_void_p, c_int, c_void_p)
_opsig_explicit_exchange_t = CFUNCTYPE(None, c_void_p, c_void_p, c_int, c_void_p)
_opsig_explicit_cmpxchg_t = CFUNCTYPE(c_int, c_void_p, c_void_p, c_void_p, c_int, c_int)
_opsig_explicit_test_t = CFUNCTYPE(c_int, c_void_p, c_int, c_int)
_opsig_explicit_test_modify_t = CFUNCTYPE(c_int, c_void_p, c_int, c_int)
_opsig_explicit_fetch_t = CFUNCTYPE(None, c_void_p, c_void_p, c_int, c_void_p)
_opsig_explicit_fetch_noarg_t = CFUNCTYPE(None, c_void_p, c_int, c_void_p)
_opsig_explicit_void_t = CFUNCTYPE(None, c_void_p, c_void_p, c_int)
_opsig_explicit_void_noarg_t = CFUNCTYPE(None, c_void_p, c_int)


class _OpsExplicitArithmetic(Structure):
    _fields_ = [("fp_add", _opsig_explicit_void_t),
                ("fp_sub", _opsig_explicit_void_t),
                ("fp_inc", _opsig_explicit_void_noarg_t),
                ("fp_dec", _opsig_explicit_void_noarg_t),
                ("fp_neg", _opsig_explicit_void_noarg_t),
                ("fp_fetch_add", _opsig_explicit_fetch_t),
                ("fp_fetch_sub", _opsig_explicit_fetch_t),
                ("fp_fetch_inc", _opsig_explicit_fetch_noarg_t),
                ("fp_fetch_dec", _opsig_explicit_fetch_noarg_t),
                ("fp_fetch_neg", _opsig_explicit_fetch_noarg_t)]


class _OpsExplicitBinary(Structure):
    _fields_ = [("fp_or", _opsig_explicit_void_t),
                ("fp_xor", _opsig_explicit_void_t),
                ("fp_and", _opsig_explicit_void_t),
                ("fp_not", _opsig_explicit_void_noarg_t),
                ("fp_fetch_or", _opsig_explicit_fetch_t),
                ("fp_fetch_xor", _opsig_explicit_fetch_t),
                ("fp_fetch_and", _opsig_explicit_fetch_t),
                ("fp_fetch_not", _opsig_explicit_fetch_noarg_t)]


class _OpsExplicitBitwise(Structure):
    _fields_ = [("fp_test", _opsig_explicit_test_t),
                ("fp_test_compl", _opsig_explicit_test_modify_t),
                ("fp_test_set", _opsig_explicit_test_modify_t),
                ("fp_test_reset", _opsig_explicit_test_modify_t)]


class _OpsExplicitXchg(Structure):
    _fields_ = [("fp_exchange", _opsig_explicit_exchange_t),
                ("fp_cmpxchg_weak", _opsig_explicit_cmpxchg_t),
                ("fp_cmpxchg_strong", _opsig_explicit_cmpxchg_t)]


class Ops(Structure):
    _fields_ = [("fp_store", _opsig_explicit_store_t),
                ("fp_load", _opsig_explicit_load_t),
                ("xchg_ops", _OpsExplicitXchg),
                ("bitwise_ops", _OpsExplicitBitwise),
                ("binary_ops", _OpsExplicitBinary),
                ("signed_ops", _OpsExplicitArithmetic),
                ("unsigned_ops", _OpsExplicitArithmetic)]


class Alignment(Structure):
    _fields_ = [("recommended", c_size_t),
                ("minimum", c_size_t),
                ("size_within", c_size_t)]


class Patomic:

    class _PatomicExplicit(Structure):
        _fields_ = [("ops", Ops),
                    ("align", Alignment)]

    _lib = None

    @staticmethod
    def _get_lib():
        if Patomic._lib is None:
            # get lib path
            path = pathlib.Path(__file__).parent.parent.resolve()
            path = path.joinpath("_clib")
            possible_paths = sorted(path.glob("*patomic*"))
            if not possible_paths:
                raise FileNotFoundError("Could not find patomic lib in atomics._clib")
            path = possible_paths[-1]
            # setup lib
            lib = cdll.LoadLibrary(str(path))
            lib.patomic_create_explicit.restype = Patomic._PatomicExplicit
            lib.patomic_create_explicit.argtypes = [c_size_t, c_int, c_int]
            lib.patomic_nonnull_ops_count_explicit.restype = c_int
            lib.patomic_nonnull_ops_count_explicit.argtypes = [POINTER(Ops)]
            # assign to static member
            Patomic._lib = lib
        return Patomic._lib

    @staticmethod
    def _create_explicit(width: int) -> _PatomicExplicit:
        char_bit = 8
        if width < 0:
            raise ValueError("Negative width")
        elif width.bit_length() > (sizeof(c_size_t) * char_bit):
            raise OverflowError(width, "Value would overflow size_t")
        return Patomic._get_lib().patomic_create_explicit(width, 0, 0)

    @staticmethod
    def ops(width: int) -> Ops:
        pae = Patomic._create_explicit(width)
        return pae.ops

    @staticmethod
    def alignment(width: int) -> Alignment:
        pae = Patomic._create_explicit(width)
        return pae.align

    @staticmethod
    def count_nonnull_ops(ops: Ops, *, readonly: bool) -> int:
        res = 0
        if readonly:
            # only current non-modifying ops
            res += bool(ops.fp_load)
            res += bool(ops.bitwise_ops.fp_test)
        else:
            res = Patomic._get_lib().patomic_nonnull_ops_count_explicit(byref(ops))
        return res



from ...enums import MemoryOrder, OpType
from ...exceptions import MemoryOrderError, UnsupportedOperationException
from ...pybuffer import PyBuffer

from ..core import AtomicCore

from .cmpxchg import CmpxchgResult

from typing import Optional, Tuple


class _ImplByteOperationsMixin:

    _core: AtomicCore

    def _impl_cmpxchg(self, optype: OpType, expected: bytes, desired: bytes,
                      succ: MemoryOrder, fail: MemoryOrder) -> CmpxchgResult[bytes]:
        assert ("CMPXCHG" in optype.name)
        # check support
        fp = self._core.get_op_func(optype)
        if fp is None:
            raise UnsupportedOperationException(optype, self._core.width, readonly=self._core.readonly)
        # validate inputs
        elif len(expected) != self._core.width:
            raise ValueError("'expected' object length does not match width.")
        elif len(desired) != self._core.width:
            raise ValueError("'desired' object length does not match width.")
        elif not fail.is_valid_fail_order(succ):
            raise MemoryOrderError(optype, fail, is_fail=True)
        # perform operation
        exp_mut = bytes(expected)  # make a copy so we can modify it
        des_mut = bytes(desired)  # make a copy so we can modify it
        with PyBuffer(exp_mut, writeable=True, force=True) as exp_buf:
            with PyBuffer(des_mut, writeable=True, force=True) as des_buf:
                # modifying exp and des contents directly is fine in this case
                ok = fp(self._core.address, exp_buf.address, des_buf.address, succ.value, fail.value)
        return CmpxchgResult(bool(ok), exp_mut)

    def _impl_bit_test(self, optype: OpType, index: int, order: MemoryOrder) -> bool:
        assert ("BIT_TEST" in optype.name)
        # check support
        fp = self._core.get_op_func(optype)
        if fp is None:
            raise UnsupportedOperationException(optype, self._core.width, readonly=self._core.readonly)
        # validate input
        elif index < 0 or index >= (self._core.width * 8):  # CHAR_BIT == 8
            raise ValueError("'index' value out of range.")
        # perform operation
        return bool(fp(self._core.address, index, order.value))

    def _impl_bin_ari(self, optype: OpType, value: Optional[bytes],
                      order: MemoryOrder) -> Optional[bytes]:
        assert (optype.value >= OpType.OR.value)
        # check support
        fp = self._core.get_op_func(optype)
        if fp is None:
            raise UnsupportedOperationException(optype, self._core.width, readonly=self._core.readonly)
        # validate input
        elif value is not None and len(value) != self._core.width:
            raise ValueError("'value' object length does not match width.")
        # setup args list
        args = [self._core.address]
        bufs = []
        # value param
        if value is not None:
            val_buf = PyBuffer(value, writeable=False)
            args.append(val_buf.address)
            bufs.append(val_buf)
        # memory order param
        args.append(order.value)
        # result "param"
        result = None
        if "FETCH" in optype.name:
            result = bytes(self._core.width)
            res_buf = PyBuffer(result, writeable=True, force=True)
            # modifying result contents directly is fine in this case
            args.append(res_buf.address)
            bufs.append(res_buf)
        # perform operation
        fp(*args)
        # release buffers and return
        for buf in bufs:
            buf.release()
        return result


class ByteOperationsMixin(_ImplByteOperationsMixin):

    _core: AtomicCore

    def store(self, desired: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        # check support
        fp = self._core.get_op_func(OpType.STORE)
        if fp is None:
            raise UnsupportedOperationException(OpType.STORE, self._core.width, readonly=self._core.readonly)
        # validate inputs
        elif not order.is_valid_store_order():
            raise MemoryOrderError(OpType.STORE, order, is_fail=False)
        elif len(desired) != self._core.width:
            raise ValueError("'desired' object length does not match width.")
        # perform operation
        with PyBuffer(desired, writeable=False) as des_buf:
            fp(self._core.address, des_buf.address, order.value)

    def load(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        # check support
        fp = self._core.get_op_func(OpType.LOAD)
        if fp is None:
            raise UnsupportedOperationException(OpType.LOAD, self._core.width, readonly=self._core.readonly)
        # validate input
        elif not order.is_valid_load_order():
            raise MemoryOrderError(OpType.LOAD, order, is_fail=False)
        # perform operation
        result = bytes(self._core.width)
        with PyBuffer(result, writeable=True, force=True) as res_buf:
            # modifying result contents directly is fine in this case
            fp(self._core.address, order.value, res_buf.address)
        return result

    def exchange(self, desired: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        # check support
        fp = self._core.get_op_func(OpType.EXCHANGE)
        if fp is None:
            raise UnsupportedOperationException(OpType.EXCHANGE, self._core.width, readonly=self._core.readonly)
        # validate input
        elif len(desired) != self._core.width:
            raise ValueError("'desired' object length does not match width.")
        # perform operation
        result = bytes(self._core.width)
        with PyBuffer(result, writeable=True, force=True) as res_buf:
            # modifying result contents directly is fine in this case
            with PyBuffer(desired, writeable=False) as des_buf:
                fp(self._core.address, des_buf.address, order.value, res_buf.address)
        return result

    def cmpxchg_weak(self, expected: bytes, desired: bytes,
                     succ: MemoryOrder = MemoryOrder.SEQ_CST,
                     fail: MemoryOrder = MemoryOrder.SEQ_CST) -> CmpxchgResult[bytes]:
        return self._impl_cmpxchg(OpType.CMPXCHG_WEAK, expected, desired, succ, fail)

    def cmpxchg_strong(self, expected: bytes, desired: bytes,
                       succ: MemoryOrder = MemoryOrder.SEQ_CST,
                       fail: MemoryOrder = MemoryOrder.SEQ_CST) -> CmpxchgResult[bytes]:
        return self._impl_cmpxchg(OpType.CMPXCHG_STRONG, expected, desired, succ, fail)

    def bit_test(self, index: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bool:
        if not order.is_valid_store_order():
            raise MemoryOrderError(OpType.BIT_TEST, order, is_fail=False)
        return self._impl_bit_test(OpType.BIT_TEST, index, order)

    def bit_test_compl(self, index: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bool:
        return self._impl_bit_test(OpType.BIT_TEST_COMPL, index, order)

    def bit_test_set(self, index: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bool:
        return self._impl_bit_test(OpType.BIT_TEST_SET, index, order)

    def bit_test_reset(self, index: int, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bool:
        return self._impl_bit_test(OpType.BIT_TEST_RESET, index, order)

    def bin_or(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.OR, value, order)

    def bin_xor(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.XOR, value, order)

    def bin_and(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.AND, value, order)

    def bin_not(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> None:
        return self._impl_bin_ari(OpType.NOT, None, order)

    def bin_fetch_or(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        return self._impl_bin_ari(OpType.FETCH_OR, value, order)

    def bin_fetch_xor(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        return self._impl_bin_ari(OpType.FETCH_XOR, value, order)

    def bin_fetch_and(self, value: bytes, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        return self._impl_bin_ari(OpType.FETCH_AND, value, order)

    def bin_fetch_not(self, order: MemoryOrder = MemoryOrder.SEQ_CST) -> bytes:
        return self._impl_bin_ari(OpType.FETCH_NOT, None, order)



from ..enums import OpType
from ..patomic import Ops
from ..pybuffer import PyBuffer

from typing import Callable, Dict, Optional


class AtomicCore:

    def __init__(self, buffer: PyBuffer, ops: Ops, *, is_integral: bool, is_signed: bool):
        # check if object has been initialised
        if hasattr(self, "_buffer"):
            raise ValueError("Core object cannot be re-initialised.")
        # setup members
        self._buffer: PyBuffer = buffer
        self._ops: Ops = ops
        self._is_integral: bool = is_integral
        self._is_signed: bool = is_signed
        self._supported: Dict[OpType, Callable] = self._get_supported_ops_map()

    def __enter__(self):
        self._assert_not_released()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __del__(self):
        self.release()

    def __bool__(self):
        return not self._released

    def release(self) -> None:
        # this may be called in __del__ if exception is raised in __init__
        # we cannot rely on any attributes existing
        if hasattr(self, "_buffer"):
            self._buffer.release()

    @property
    def _released(self) -> bool:
        return not bool(self._buffer)

    def _assert_not_released(self) -> None:
        if self._released:
            msg = f"Operation forbidden on released {self.__class__.__name__} object."
            raise ValueError(msg)

    def get_op_func(self, optype: OpType) -> Optional[Callable]:
        self._assert_not_released()
        return self._supported.get(optype)

    def _get_supported_ops_map(self) -> Dict[OpType, Callable]:
        self._assert_not_released()
        ots: Dict[OpType, Callable] = {}
        # loop through all possible ops
        for ot in OpType:
            # ignore arithmetic ops if not integral
            if not self._is_integral and ot >= OpType.ADD:
                continue
            # ignore mutating ops if readonly
            if self._buffer.readonly and ot not in (OpType.LOAD, OpType.BIT_TEST):
                continue
            # get op category (e.g. _ops for load, _ops.xchg_ops for exchange, etc...)
            cat = self._ops
            if ot.cname is not None:
                s_type = "signed" if self._is_signed else "unsigned"
                cname = ot.cname.replace("arithmetic", s_type)
                cat = getattr(cat, cname)
            # get op function pointer
            fp = getattr(cat, ot.fname)
            # add to ots if supported (i.e. fp is not NULL)
            if fp:
                ots[ot] = fp
        # return supported ops
        return ots

    @property
    def address(self) -> int:
        self._assert_not_released()
        return self._buffer.address

    @property
    def width(self) -> int:
        self._assert_not_released()
        return self._buffer.width

    @property
    def readonly(self) -> bool:
        self._assert_not_released()
        return self._buffer.readonly

    @property
    def signed(self) -> bool:
        self._assert_not_released()
        return self._is_signed

    @property
    def ops_supported(self) -> [OpType]:
        self._assert_not_released()
        return sorted(list(self._supported.keys()))



from typing import Generic, TypeVar


T = TypeVar('T', bytes, int)


class CmpxchgResult(Generic[T]):

    def __init__(self, success: bool, expected: T):
        self.success: bool = success
        self.expected: T = expected

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(success={self.success}, " \
               f"expected={self.expected})"

    def __bool__(self) -> bool:
        return self.success

    def __iter__(self):
        return iter((self.success, self.expected))



from ._impl.atomic.base import AtomicView
from ._impl.atomic.baseint import AtomicIntegralView
from ._impl.atomic.bytes import AtomicBytesView
from ._impl.atomic.int import AtomicIntView, AtomicUintView



from ._impl.alignment import Alignment
from ._impl.enums import MemoryOrder, OpType

from ._impl.atomic.funcs import atomic, atomicview

from ._impl.atomic.mixins.cmpxchg import CmpxchgResult
from ._impl.atomic.mixins.types import ANY, INTEGRAL, BYTES, INT, UINT

__all__ = [
    "atomic", "atomicview",
    "ANY", "INTEGRAL", "BYTES", "INT", "UINT",
    "Alignment", "CmpxchgResult",
    "MemoryOrder", "OpType",
]



from .enums import MemoryOrder, OpType


class AlignmentError(Exception):

    def __init__(self, width: int, address: int, *, using_recommended: bool):
        self.width: int = width
        self.address: int = address
        self.using_recommended: int = using_recommended
        u_type = "recommended" if using_recommended else "minimum"
        message = f"The alignment of address {hex(address)} does not meet " \
                  f"the {u_type} alignment requirements for atomic operations " \
                  f"on objects with a width of {width}."
        super().__init__(message)

    def __repr__(self):
        r = f"{self.__class__.__name__}(width={self.width}, address={self.address}" \
            f", using_recommended={self.using_recommended})"
        return r


class MemoryOrderError(Exception):

    def __init__(self, optype: OpType, order: MemoryOrder, *, is_fail: bool):
        self.optype: OpType = optype
        self.order: MemoryOrder = order
        self.is_fail: bool = is_fail
        f_type = "fail " if is_fail else " "
        v_type = "an" if optype.name.lower()[0] in "aeiou" else "a"
        message = f"{order.name} is not a valid {f_type}memory order for " \
                  f"{v_type} {optype.name} operation."
        super().__init__(message)

    def __repr__(self):
        r = f"{self.__class__.__name__}(optype=OpType.{self.optype.name}, order=" \
            f"MemoryOrder.{self.order.name}, is_fail={self.is_fail})"
        return r


class UnsupportedWidthException(Exception):

    def __init__(self, width: int, *, readonly: bool):
        self.width: int = width
        self.readonly: bool = readonly
        r_type = " readonly " if readonly else " "
        message = f"No operations are supported on{r_type}objects with a width " \
                  f"of {width}."
        super().__init__(message)

    def __repr__(self):
        r = f"{self.__class__.__name__}(width={self.width}, readonly={self.readonly})"
        return r


class UnsupportedOperationException(Exception):

    def __init__(self, optype: OpType, width: int, *, readonly: bool):
        self.optype: OpType = optype
        self.width: int = width
        self.readonly: bool = readonly
        r_type = "readonly " if readonly else " "
        message = f"Operation {optype.name} is not supported on {r_type}objects " \
                  f"with a width of {width}."
        super().__init__(message)

    def __repr__(self):
        r = f"{self.__class__.__name__}(optype=OpType.{self.optype.name}, " \
            f"width={self.width}, readonly={self.readonly})"
        return r



from ._impl.atomic.base import AtomicViewContext
from ._impl.atomic.baseint import AtomicIntegralViewContext
from ._impl.atomic.bytes import AtomicBytesViewContext
from ._impl.atomic.int import AtomicIntViewContext, AtomicUintViewContext



from .base import Atomic, AtomicViewContext
from .baseint import AtomicIntegral, AtomicIntegralViewContext
from .bytes import AtomicBytes, AtomicBytesViewContext
from .int import AtomicInt, AtomicIntViewContext
from .int import AtomicUint, AtomicUintViewContext

from .mixins.types import ANY, INTEGRAL, BYTES, INT, UINT

from typing import overload, Type, Union


ATUnion = Union[
    ANY,
    INTEGRAL,
    BYTES,
    INT,
    UINT
]

AUnion = Union[
    Atomic,
    AtomicIntegral,
    AtomicBytes,
    AtomicInt,
    AtomicUint
]

AVCUnion = Union[
    AtomicViewContext,
    AtomicIntegralViewContext,
    AtomicBytesViewContext,
    AtomicIntViewContext,
    AtomicUintViewContext
]


@overload
def atomic(width: int, atype: Type[INT], **kwargs) -> AtomicInt:
    ...


@overload
def atomic(width: int, atype: Type[UINT], **kwargs) -> AtomicUint:
    ...


@overload
def atomic(width: int, atype: Type[BYTES], **kwargs) -> AtomicBytes:
    ...


@overload
def atomic(width: int, atype: Type[INTEGRAL], **kwargs) -> AtomicIntegral:
    ...


@overload
def atomic(width: int, atype: Type[ANY], **kwargs) -> Atomic:
    ...


def atomic(width: int, atype: Type[ATUnion], **kwargs) -> AUnion:
    if atype is INT:
        return AtomicInt(width=width)
    elif atype is UINT:
        return AtomicUint(width=width)
    elif atype is BYTES:
        return AtomicBytes(width=width)
    elif atype is INTEGRAL:
        return AtomicIntegral(width=width, **kwargs)
    elif atype is ANY:
        return Atomic(width=width, **kwargs)
    else:
        msg = "Type parameter 'atype' must be one of [ANY, INTEGRAL, BYTES, INT, UINT]."
        raise TypeError(msg)


@overload
def atomicview(buffer, atype: Type[INT], **kwargs) -> AtomicIntViewContext:
    ...


@overload
def atomicview(buffer, atype: Type[UINT], **kwargs) -> AtomicUintViewContext:
    ...


@overload
def atomicview(buffer, atype: Type[BYTES], **kwargs) -> AtomicBytesViewContext:
    ...


@overload
def atomicview(buffer, atype: Type[INTEGRAL], **kwargs) -> AtomicIntegralViewContext:
    ...


@overload
def atomicview(buffer, atype: Type[ANY], **kwargs) -> AtomicViewContext:
    ...


def atomicview(buffer, atype: Type[ATUnion], **kwargs) -> AVCUnion:
    if atype is INT:
        return AtomicIntViewContext(buffer=buffer)
    elif atype is UINT:
        return AtomicUintViewContext(buffer=buffer)
    elif atype is BYTES:
        return AtomicBytesViewContext(buffer=buffer)
    elif atype is INTEGRAL:
        return AtomicIntegralViewContext(buffer=buffer, **kwargs)
    elif atype is ANY:
        return AtomicViewContext(buffer=buffer, **kwargs)
    else:
        msg = "Type parameter 'atype' must be one of [ANY, INTEGRAL, BYTES, INT, UINT]."
        raise TypeError(msg)



from setuptools import setup, find_packages, Command, Extension
from setuptools.command import build_py
from wheel.bdist_wheel import bdist_wheel

import configparser
import git
import logging
import os
import pathlib
import subprocess
import sys
import tempfile

from typing import List, Optional


here = pathlib.Path(__file__).parent.resolve()


class BdistWheelCommand(bdist_wheel):

    @staticmethod
    def _supported_versions() -> List[str]:
        """Parses setup.cfg to return equiv of ['36', '37', ...]"""
        cfg = configparser.ConfigParser()
        cfg.read("setup.cfg")
        classifiers = cfg["metadata"]["classifiers"].split("\n")
        prefix = "Programming Language :: Python ::"
        versions = []
        for c in classifiers:
            if c.startswith(prefix) and "." in c:
                # prefix X.Y
                v = c.strip(prefix).strip().replace(".", "")
                versions.append(v)
        return versions

    @staticmethod
    def _in_ci() -> bool:
        return bool(os.environ.get("CI"))

    def get_tag(self):
        python, abi, plat = super().get_tag()
        abi = "none"
        if self._in_ci():
            python = "py" + ".py".join(self._supported_versions())
        return python, abi, plat


class BuildPatomicCommand(Command):

    logger: logging.Logger

    git_url: str
    git_tag: Optional[str]
    dest_dir: pathlib.Path
    build_type: str
    force_replace: bool
    cc_path: Optional[pathlib.Path]
    cc_standard: int
    cmake_args: str
    log_level: int

    description = "Clone and build patomic shared library"
    user_options = [
        ("git-url=", 'u', "[str] URL to patomic git repo"),
        ("git-tag=", 't', "[str] Tag of commit to use from git repo"),
        ("dest-dir=", 'd', "[str] Directory to place built patomic shared library"),
        ("build-type=", 'b', "[str] CMake build type"),
        ("force-replace=", 'f', "[bool] Replace existing patomic file(s)"),
        ("cc-path=", 'c', "[str] C compiler path"),
        ("cc-standard=", 's', "[int] ISO C standards version"),
        ("cmake-args=", 'a', "[str] Opaque string appended to cmake config command"),
        ("log-level=", 'l', "[int] Log level")
    ]

    def obtain_env_options(self) -> None:
        for name, _, _ in self.user_options:
            attr_name = name.replace("-", "_")[:-1]
            env_name = f"BUILD_PATOMIC_{attr_name.upper()}"
            env = os.environ.get(env_name)
            if env:
                setattr(self, attr_name, env)

    def initialize_options(self) -> None:
        self.git_url = "https://github.com/doodspav/patomic"
        self.git_tag = None
        self.dest_dir = here / "src" / "atomics" / "_clib"
        self.build_type = "RelWithDebInfo"
        self.force_replace = False
        self.cc_path = None
        self.cc_standard = 11
        self.cmake_args = ""
        self.log_level = logging.DEBUG if self._in_ci() else logging.INFO
        # override defaults with env here
        # command line takes precedence over env
        self.obtain_env_options()

    def finalize_options(self) -> None:
        # coerce types
        self.dest_dir = pathlib.Path(self.dest_dir).resolve()
        self.cc_standard = int(self.cc_standard)
        # fix bool
        if isinstance(self.force_replace, str):
            sfr = self.force_replace.lower().strip()
            if sfr.isnumeric():
                self.force_replace = bool(int(self.force_replace))
            else:
                self.force_replace = (sfr == "true")
        # coerce compiler values
        if self.cc_path is not None:
            self.cc_path = pathlib.Path(self.cc_path)
        if self.cc_standard in [89, 95]:
            self.cc_standard = 90
        # setup logger
        self.log_level = int(self.log_level)
        self._init_logger()
        self._log_options()

    def _init_logger(self) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.log_level)
        # setup formatting
        sh = logging.StreamHandler()
        template = "[build_patomic] [{levelname}] [{funcName}] {message}"
        fmt = logging.Formatter(template, style="{")
        sh.setFormatter(fmt)
        self.logger.addHandler(sh)

    def _log_options(self) -> None:
        opts = {}
        for name, _, _ in self.user_options:
            attr = name.replace("-", "_")[:-1]
            opts[attr] = getattr(self, attr)
        self.logger.debug(str(opts))

    @staticmethod
    def _in_ci() -> bool:
        """Checks if running in CI (via environment variable)"""
        return bool(os.environ.get("CI"))

    @staticmethod
    def _cibw_check_win32_x86() -> bool:
        """Checks if CIBW is building for win32-x86"""
        env = os.environ.get("CIBW_MC_NAME")
        return type(env) is str and env.lower() == "win32-x86"

    def get_patomic_libs(self, dir_path: pathlib.Path) -> [pathlib.Path]:
        """Returns a list of patomic shared library files found in dir_path"""
        assert dir_path.is_dir()
        exts = [".dll", ".dylib", ".so"]
        self.logger.debug(f"all paths in {str(dir_path)}: {list(dir_path.iterdir())}")
        files = [p for p in dir_path.iterdir() if p.is_file()]
        self.logger.debug(f"files: {files}")
        libs = [f for f in files if any(map(str(f).__contains__, exts))]
        self.logger.debug(f"libs: {libs}")
        patomic_libs = [f for f in libs if "patomic" in str(f)]
        return patomic_libs

    def clone_patomic(self, clone_into: pathlib.Path) -> None:
        """Clones git_url into clone_to directory"""
        assert clone_into.is_dir()
        # clone default branch
        repo = git.Repo.clone_from(url=self.git_url, to_path=str(clone_into))
        self.logger.debug(f"Cloned patomic repo into: {str(clone_into)}")
        if self.git_tag:
            self.logger.info(f"Checkout out user provided tag: {self.git_tag}")
            repo.git.checkout(self.git_tag)
        # switch to devel branch if tag or default branch isn't populated
        self.logger.info("Switching to devel branch")
        if not (clone_into / "src").is_dir():
            repo.git.checkout("devel")

    def build_patomic(self, repo_dir: pathlib.Path) -> pathlib.Path:
        """Builds patomic in repo_dir and returns shared library file path"""
        assert repo_dir.is_dir()
        use_shell = (os.name == "nt")
        fd_out = sys.stdout if self.logger.level <= logging.DEBUG else subprocess.DEVNULL
        # configure build
        os.mkdir(str(repo_dir / "build"))
        cmd_config = [
            "cmake", "-S", str(repo_dir), "-B", str(repo_dir / "build"),
            f"-DCMAKE_BUILD_TYPE={self.build_type}",
            f"-DCMAKE_C_STANDARD={self.cc_standard}",
            "-DBUILD_SHARED_LIBS=ON"
        ]
        if self.cc_path:
            self.logger.info(f"C compiler set to {self.cc_path}")
            cmd_config.append(f"-DCMAKE_C_COMPILER={str(self.cc_path)}")
        if self._cibw_check_win32_x86():
            self.logger.info(f"Running under win32-x86 on CIBW - using '-A Win32'")
            cmd_config.append("-A Win32")
        subprocess.check_call(cmd_config, stdout=fd_out, shell=use_shell)
        self.logger.debug("Configured CMake for patomic")
        # build
        build_path = repo_dir / "build"
        cmd_build = ["cmake", "--build", str(build_path), "--config", self.build_type]
        subprocess.check_call(cmd_build, stdout=fd_out, shell=use_shell)
        self.logger.debug(f"Built patomic in: {str(build_path)}")
        # check if using multi-config generator
        if (build_path / self.build_type).is_dir():
            build_path /= self.build_type
        # return most specific lib path (e.g. .so.x.y.z not .so)
        # done by length to support both .so.x.y.z and .x.y.z.dylib
        lib_paths = self.get_patomic_libs(build_path)
        lib_paths.sort(reverse=True, key=lambda p: len(str(p)))
        assert lib_paths, "Should not be empty if CMake succeeds"
        self.logger.debug(f"Chosen patomic lib path: {str(lib_paths[0])}")
        return lib_paths[0]

    def run(self):
        self.logger.debug(f"Contents of dest_dir: {list(self.dest_dir.iterdir())}")
        existing_lib_paths = self.get_patomic_libs(self.dest_dir)
        self.logger.debug(f"Existing lib paths: {existing_lib_paths}")
        # check if we can replace libs
        if existing_lib_paths and not self.force_replace:
            self.logger.info("Skipping; library already build")
            return
        # create shared library and copy over
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                # clone repo and build
                repo_path = pathlib.Path(temp_dir)
                self.logger.info("Cloning repo")
                self.clone_patomic(repo_path)
                self.logger.info("Building shared library")
                lib_path = self.build_patomic(repo_path)
                self.logger.info("Built shared library successfully")
                # copy stuff over
                if self.force_replace and existing_lib_paths:
                    self.logger.info("Removing existing shared library file(s)")
                    for elp in existing_lib_paths:
                        try:
                            elp.unlink(missing_ok=True)
                        except PermissionError:
                            self.logger.info(f"Could not remove file {str(elp)}")
                self.logger.info(f"Copying over shared library file to: {str(self.dest_dir)}")
                os.rename(str(lib_path), str(self.dest_dir / lib_path.name))
                # log result
                self.logger.info("Copied over file successfully")
                self.logger.debug(f"Files in {str(self.dest_dir)}: {list(self.dest_dir.iterdir())}")
        except PermissionError:
            self.logger.info(f"Could not close temporary directory")


class BuildPyCommand(build_py.build_py):

    def run(self):
        self.run_command("build_patomic")
        build_py.build_py.run(self)


# setup dummy extension to force non-pure python build
# CIBW won't build pure python projects
with open("dummy.c", "w") as f:
    f.truncate()
    f.write("/* temporary file - safe to delete */\n")
    f.write("extern int PyInit_dummy(void) { return 0; }\n")


setup(
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"atomics": ["_clib/*"]},
    cmdclass={
        "bdist_wheel": BdistWheelCommand,
        "build_patomic": BuildPatomicCommand,
        "build_py": BuildPyCommand
    },
    ext_modules=[Extension(name="dummy", sources=["dummy.c"])]
)


# cleanup extension
os.remove("dummy.c")



from ..alignment import Alignment
from ..exceptions import AlignmentError, UnsupportedWidthException
from ..patomic import Patomic
from ..pybuffer import PyBuffer

from .core import AtomicCore

from .mixins.types import ANY


class Atomic(ANY):

    def __init__(self, *, width: int, is_integral: bool, is_signed: bool):
        # check if object has been initialised
        if hasattr(self, "_core"):
            raise ValueError("Atomic object cannot be re-initialised.")
        # check type
        if not isinstance(width, int):
            raise TypeError("Keyword argument 'width' must have type 'int'.")
        # check ops are available
        p = Patomic()
        ops = p.ops(width)
        if p.count_nonnull_ops(ops, readonly=False) == 0:
            raise UnsupportedWidthException(width, readonly=False)
        # check alignment of buffer
        pybuf = PyBuffer(bytearray(width), writeable=True)
        align = Alignment(width)
        if not align.is_valid_recommended(pybuf.obj):
            raise AlignmentError(width, pybuf.address, using_recommended=True)
        # create core
        self._core = AtomicCore(pybuf, ops, is_integral=is_integral, is_signed=is_signed)

    def __del__(self):
        if hasattr(self, "_core"):
            self._core.release()


class AtomicView(ANY):

    def __init__(self, core: AtomicCore):
        # check if object has been initialised
        if hasattr(self, "_core"):
            raise ValueError("AtomicView object cannot be re-initialised.")
        # check type
        if not isinstance(core, AtomicCore):
            raise TypeError("Positional argument 'core' must have type 'AtomicCore'.")
        # store core
        self._core = core

    def __del__(self):
        if hasattr(self, "_core"):
            self._core.release()


class AtomicViewContext:

    def __init__(self, *, buffer, is_integral: bool, is_signed: bool):
        # check if object has been initialised
        if hasattr(self, "_core"):
            raise ValueError("AtomicViewContext object cannot be re-initialised.")
        # check and deal with buffer
        pybuf = None
        try:
            with memoryview(buffer) as view:
                pybuf = PyBuffer(buffer, writeable=(not view.readonly))
        except TypeError:
            pass
        # check for TypeError; raise outside exception handler for nicer error message
        if pybuf is None:
            em = "Keyword argument 'buffer' must support the buffer protocol."
            raise TypeError(em)
        # check ops are available
        p = Patomic()
        ops = p.ops(pybuf.width)
        if p.count_nonnull_ops(ops, readonly=pybuf.readonly) == 0:
            # pybuf MUST be released before function exit
            width, ro = pybuf.width, pybuf.readonly
            pybuf.release()
            raise UnsupportedWidthException(width, readonly=ro)
        # check alignment of buffer
        align = Alignment(pybuf.width)
        if not align.is_valid_recommended(pybuf.obj):
            # pybuf MUST be released before function exit
            width, addr = pybuf.width, pybuf.address
            pybuf.release()
            raise AlignmentError(width, addr, using_recommended=True)
        # create core
        self._core = AtomicCore(pybuf, ops, is_integral=is_integral, is_signed=is_signed)
        self._entered = False
        self._exited = False

    def __enter__(self) -> AtomicView:
        self._assert_enter_preconditions()
        self._entered = True
        return AtomicView(self._core)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exited = True
        self.release()

    def __del__(self):
        self.release()

    def release(self) -> None:
        if hasattr(self, "_core"):
            if self._entered and not self._exited:
                raise ValueError("Cannot call 'release' while context is open.")
            else:
                self._core.release()

    def _assert_enter_preconditions(self) -> None:
        if self._entered:
            raise ValueError("Cannot open context multiple times.")
        elif not self._core:
            raise ValueError("Cannot open context after calling 'release'.")



import enum
from typing import Optional


class MemoryOrder(enum.IntEnum):

    RELAXED = 0
    # CONSUME = 1
    ACQUIRE = 2
    RELEASE = 3
    ACQ_REL = 4
    SEQ_CST = 5

    def is_valid_store_order(self) -> bool:
        return self.value not in (1, 2, 4)  # CONSUME, ACQUIRE, ACQ_REL

    def is_valid_load_order(self) -> bool:
        return self.value not in (3, 4)  # RELEASE, ACQ_REL

    def is_valid_fail_order(self, succ: "MemoryOrder") -> bool:
        return (self.value <= succ.value) and self.is_valid_load_order()


class OpType(enum.IntEnum):

    def __init__(self, value):
        fname: str = self.name.lower()
        # get function name
        if fname.startswith("bit"):
            fname = fname[len("bit_"):]
        fname = "fp_" + fname
        # get category name
        cname: Optional[str] = None
        if "xch" in fname:
            cname = "xchg_ops"
        elif "test" in fname:
            cname = "bitwise_ops"
        elif fname.endswith(("or", "xor", "and", "not")):
            cname = "binary_ops"
        elif fname not in ("fp_store", "fp_load"):
            cname = "arithmetic_ops"
        # assign to self
        self.fname = fname
        self.cname = cname

    # base (category=None)
    STORE = enum.auto()
    LOAD = enum.auto()

    # xchg
    EXCHANGE = enum.auto()
    CMPXCHG_WEAK = enum.auto()
    CMPXCHG_STRONG = enum.auto()

    # bit-wise
    BIT_TEST = enum.auto()
    BIT_TEST_COMPL = enum.auto()
    BIT_TEST_SET = enum.auto()
    BIT_TEST_RESET = enum.auto()

    # binary
    OR = enum.auto()
    XOR = enum.auto()
    AND = enum.auto()
    NOT = enum.auto()
    FETCH_OR = enum.auto()
    FETCH_XOR = enum.auto()
    FETCH_AND = enum.auto()
    FETCH_NOT = enum.auto()

    # arithmetic
    ADD = enum.auto()
    SUB = enum.auto()
    INC = enum.auto()
    DEC = enum.auto()
    NEG = enum.auto()
    FETCH_ADD = enum.auto()
    FETCH_SUB = enum.auto()
    FETCH_INC = enum.auto()
    FETCH_DEC = enum.auto()
    FETCH_NEG = enum.auto()



from .byteops import ByteOperationsMixin
from .intops import IntegralOperationsMixin
from .properties import BasePropertiesMixin, BytePropertiesMixin, IntegralPropertiesMixin


class ANY(BasePropertiesMixin):
    pass


class INTEGRAL(ANY, IntegralOperationsMixin, IntegralPropertiesMixin):
    pass


class BYTES(ANY, ByteOperationsMixin, BytePropertiesMixin):
    pass


class INT(INTEGRAL):
    pass


class UINT(INTEGRAL):
    pass
